from winappdbg import *
import winappdbg
import time
import os
import sys
import subprocess
import random
from winappdbg.textio import CrashDump

## Generic in-memory-fuzzer
## If you find bugs, email apillai@cigital.com

DETACHED_PROCESS = 0x00000008 #CreateProcess Flag


first_time=True
fuzzing_started=False
saved_context={}
function_address = 0
return_address = 0
address=0


def saveMemorySnapshot(event):
    global memory_snapshot
    p = event.get_process()
    memory_snapshot = p.take_memory_snapshot()

def restoreMemorySnapshot(event):
     global memory_snapshot
     p = event.get_process()
     p.restore_memory_snapshot(memory_snapshot,bSkipMappedFiles=True)

   
def saveContext(event):
    global saved_context
    
    t = event.get_thread()
        
    #save registers,EFlags etc
    saved_context = t.get_context()
    


def restoreContext(event):
    global saved_context
    
    t = event.get_thread()
    
    
    #restore the memory state
    t.set_context(saved_context)
    
    

def on_function_entry(event):
    
    global fuzzing_started
    global first_time
    
    #print "on_function_entry"
    t = event.get_thread()
    
    if(first_time):

        print "Disassembly of chosen function "

        eip  = t.get_pc()

        print CrashDump.dump_code(t.disassemble( eip, 64 ))
            
        raw_input("Press any key to continue. You can quit by pressing Ctrl+C.")

        first_time=False
        
    #read the stack for arguments
    params  = t.read_stack_dwords(PARAM_COUNT+1,0)
    
    if(fuzzing_started == False):
        print "Taking memory snapshot .."
        
        saveMemorySnapshot(event)
        saveContext(event)

        #set breakpoint at the return address
        return_address = params[0]
        event.debug.break_at( event.get_pid(), return_address, on_function_exit )
        
        fuzzing_started=True
        print "Fuzzing Started"
        
    #modify arguments passed to the funtion
    modifyParams(event,params)

    

#Called when function exits
def on_function_exit(event):

    global function_address
    
    
    t = event.get_thread()


    #free memory
    freeMemory(event)

    #restore memory, registers, flags
    restoreMemorySnapshot(event)
    restoreContext(event)

    #set EIP to start of function
    t.set_pc(function_address)

    

    
class debugEventHandler( EventHandler ):
    
    def __init__(self, logfile):
        # Create the logger object.
        self.logger = Logger(logfile, True)
        
    def _post_exception(self, event):
        #Kill the process
        if event.is_last_chance():

            try:
                event.get_thread().set_pc(event.get_process().resolve_symbol('kernel32!ExitProcess') )
            except Exception:
                event.get_process().kill()
            #Start the process again
            
            if(RESTART_PROCESS==True and ATTACH==True):
                print "Restarting process"
                subprocess.Popen(PROCESS_DETAILS,creationflags=DETACHED_PROCESS)

            
            
        
    def exception( self, event):
        if event.is_last_chance():
            crash = Crash(event)
            crash.fetch_extra_data(event, 0)
            msg = ""
            msg = msg + "\n***************************************************\n"
            msg = msg + "Crash induced by: "+repr(get_fuzzstring())+"\n"
            msg = msg + "***************************************************\n"
            msg = msg + crash.fullReport(bShowNotes = True)
            self.logger.log_event(event, msg)
        self._post_exception(event)
            
    
    def load_dll( self, event ):
        global function_address
        
               
        if(len(MODULE_NAME) != 0):
            # Get the new module object
            module = event.get_module()
            if(module.match_name(MODULE_NAME)):
                print MODULE_NAME+" loaded at "+hex(module.get_base())
                if("FUNCTION_ADDRESS" in globals() and FUNCTION_ADDRESS>0):
                    function_address = FUNCTION_ADDRESS
                else:
                    function_address = module.get_base() + FUNCTION_OFFSET
                
                event.debug.break_at( event.get_pid(),function_address , on_function_entry )
            

def setBPMainModule(debug,process):
    if("FUNCTION_ADDRESS" in globals() and FUNCTION_ADDRESS>0):
        function_address = FUNCTION_ADDRESS
    else:
        function_address = process.get_image_base() + 0x1000+FUNCTION_OFFSET
    debug.break_at(process.get_pid(), function_address, on_function_entry)

def debug_app():
    global fuzzing_started
    
    event_handler = debugEventHandler(LOG_FILE)
    
    try:
        
        while(1):
       # For all processes that match the requested filename...
       
            if(ATTACH==True):
                
                    debug = Debug(event_handler)
                    for ( process, name ) in debug.system.find_processes_by_filename( DEBUG_PROCESS ):

                        #Print the process id and process name
                        pid = process.get_pid()
                        print pid, name

                        #This flag is set here again to handle process restarts
                        fuzzing_started=False
                        
                        # Attach to the process
                        debug.attach( pid )

                        #If module name is not specified
                        if(len(MODULE_NAME) == 0):
                            setBPMainModule(debug,process)
                        debug.loop()
                    time.sleep(5)
            else:
                debug=Debug(event_handler)
                print PROCESS_DETAILS
                process=debug.execv( PROCESS_DETAILS)
                #winappdbg fails when get_main_module function is called. Temporarily fixed by adding 0x1000 to find the
                #module base from image base
                if(len(MODULE_NAME) == 0):
                    setBPMainModule(debug,process)
                debug.loop()
            
                
            if(RESTART_PROCESS==False):
                break

                    
                        
        # Stop the debugger
    finally:
        debug.stop()

#Get module name from command line
if (len(sys.argv) > 2 and sys.argv[1]=="-m"):
    
    module_name = sys.argv[2]

    #Dynamically import the module
    imported_module = __import__(module_name)

    #Simulate from module import *
    for attr in dir(imported_module):
        if not attr.startswith('_'):
            globals()[attr] = getattr(imported_module, attr)

    #Start debugging
    debug_app()
    
else:
    print "Usage:\npython in_memory_fuzz.py -m <modulename>\n\nExample:python in_memory_fuzz.py -m imf_config\n\n"+\
    "Make sure the module has unique name and is in the same directory as the python file"
    

